'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var events = _interopDefault(require('events'));
var url = _interopDefault(require('url'));

'use strict';
function _interopDefault$1 (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }
var require$$0 = _interopDefault$1(events);
var EventEmitter;
var Resource$1;
({EventEmitter} = require$$0);
Resource$1 = class Resource extends EventEmitter {
  constructor(identifier) {
    if (!identifier) {
      throw TypeError('Missing identifier');
    }
    super();
    this.identifier = identifier;
  }
};
var Resource_1 = Resource$1;
var EventEmitter$2;
var Registry$2;
({EventEmitter: EventEmitter$2} = require$$0);
Registry$2 = class Registry extends EventEmitter$2 {
  constructor() {
    super();
    this.resources = {};
  }
  isRegistered(resource) {
    var ref;
    return (ref = this.resources[resource.identifier]) != null ? typeof ref.includes === "function" ? ref.includes(resource) : void 0 : void 0;
  }
  register(resource) {
    var identifier, resources;
    if (this.isRegistered(resource)) {
      return resource;
    }
    ({identifier} = resource);
    if (!(resources = this.resources[identifier])) {
      resources = this.resources[identifier] = [];
      this.emit('init', resource, this);
    }
    this.emit('register', resource, this);
    resources.push(resource);
    return () => {
      return this.unregister(resource);
    };
  }
  unregister(resource) {
    var identifier, resources;
    if (!this.isRegistered(resource)) {
      return resource;
    }
    ({identifier} = resource);
    resources = this.resources[identifier];
    this.emit('unregister', resource, this);
    resources = this.resources[identifier].filter(function(r) {
      return r !== resource;
    });
    if (resources.length === 0) {
      delete this.resources[identifier];
      return this.emit('release', resource, this);
    } else {
      return this.resources[identifier] = resources;
    }
  }
};
var Registry_1 = Registry$2;
var EventEmitter$1;
var Manager$1;
var Registry$1;
({EventEmitter: EventEmitter$1} = require$$0);
Registry$1 = Registry_1;
Manager$1 = class Manager extends EventEmitter$1 {
  constructor(registry) {
    super();
    this.registry = registry;
    if (this.registry == null) {
      this.registry = new Registry$1;
    }
    this.registry.on('init', (resource) => {
      return typeof this.init === "function" ? this.init(resource) : void 0;
    });
    this.registry.on('register', (resource) => {
      return typeof this.register === "function" ? this.register(resource) : void 0;
    });
    this.registry.on('unregister', (resource) => {
      return typeof this.unregister === "function" ? this.unregister(resource) : void 0;
    });
    this.registry.on('release', (resource) => {
      return typeof this.release === "function" ? this.release(resource) : void 0;
    });
  }
  manage(...resources) {
    resources.forEach((resource) => {
      return this.registry.register(resource);
    });
    return this;
  }
  unmanage(...resources) {
    resources.forEach((resource) => {
      return this.registry.unregister(resource);
    });
    return this;
  }
};
var Manager_1 = Manager$1;
var Manager$1$1;
var Registry$1$1;
var Resource$1$1;
Resource$1$1 = Resource_1;
Manager$1$1 = Manager_1;
Registry$1$1 = Registry_1;
var build$2 = {Resource: Resource$1$1, Registry: Registry$1$1, Manager: Manager$1$1};
var arm = build$2;

var Resource$2;
var URL;
var arm$2;
arm$2 = arm;
URL = url;
Resource$2 = class Resource extends arm$2.Resource {
  constructor(name, opts) {
    var url$$1;
    if (typeof name === 'object') {
      opts = name;
      name = opts.name;
    }
    if (opts == null) {
      opts = {};
    }
    if (!name) {
      throw Error('Missing resource name');
    }
    if (name.match(/^https?:\/\//)) {
      url$$1 = URL.parse(name);
      super(`${url$$1.host}${url$$1.pathname}`);
      this.remote = name;
    } else {
      super(name);
    }
    this.type = opts.type || 'pull';
    this.queue = opts.queue || 'oneshot';
    this.docs = opts.docs || 'all';
  }
};
var Resource_1$1 = Resource$2;

'use strict';
function lowerBound(array, value, comp) {
	let first = 0;
	let count = array.length;
	while (count > 0) {
		const step = (count / 2) | 0;
		let it = first + step;
		if (comp(array[it], value) <= 0) {
			first = ++it;
			count -= step + 1;
		} else {
			count = step;
		}
	}
	return first;
}
class PriorityQueue {
	constructor() {
		this._queue = [];
	}
	enqueue(run, opts) {
		opts = Object.assign({
			priority: 0
		}, opts);
		const element = {priority: opts.priority, run};
		if (this.size && this._queue[this.size - 1].priority >= opts.priority) {
			this._queue.push(element);
			return;
		}
		const index = lowerBound(this._queue, element, (a, b) => b.priority - a.priority);
		this._queue.splice(index, 0, element);
	}
	dequeue() {
		return this._queue.shift().run;
	}
	get size() {
		return this._queue.length;
	}
}
class PQueue$1 {
	constructor(opts) {
		opts = Object.assign({
			concurrency: Infinity,
			autoStart: true,
			queueClass: PriorityQueue
		}, opts);
		if (!(typeof opts.concurrency === 'number' && opts.concurrency >= 1)) {
			throw new TypeError(`Expected \`concurrency\` to be a number from 1 and up, got \`${opts.concurrency}\` (${typeof opts.concurrency})`);
		}
		this.queue = new opts.queueClass();
		this._queueClass = opts.queueClass;
		this._pendingCount = 0;
		this._concurrency = opts.concurrency;
		this._isPaused = opts.autoStart === false;
		this._resolveEmpty = () => {};
		this._resolveIdle = () => {};
	}
	_next() {
		this._pendingCount--;
		if (!this._isPaused && this.queue.size > 0) {
			this.queue.dequeue()();
		} else {
			this._resolveEmpty();
			if (this._pendingCount === 0) {
				this._resolveIdle();
			}
		}
	}
	add(fn, opts) {
		return new Promise((resolve, reject) => {
			const run = () => {
				this._pendingCount++;
				fn().then(
					val => {
						resolve(val);
						this._next();
					},
					err => {
						reject(err);
						this._next();
					}
				);
			};
			if (!this._isPaused && this._pendingCount < this._concurrency) {
				run();
			} else {
				this.queue.enqueue(run, opts);
			}
		});
	}
	addAll(fns, opts) {
		return Promise.all(fns.map(fn => this.add(fn, opts)));
	}
	start() {
		if (!this._isPaused) {
			return;
		}
		this._isPaused = false;
		while (this.queue.size > 0 && this._pendingCount < this._concurrency) {
			this.queue.dequeue()();
		}
	}
	pause() {
		this._isPaused = true;
	}
	clear() {
		this.queue = new this._queueClass();
	}
	onEmpty() {
		if (this.queue.size === 0) {
			return Promise.resolve();
		}
		return new Promise(resolve => {
			const existingResolve = this._resolveEmpty;
			this._resolveEmpty = () => {
				existingResolve();
				resolve();
			};
		});
	}
	onIdle() {
		if (this._pendingCount === 0) {
			return Promise.resolve();
		}
		return new Promise(resolve => {
			const existingResolve = this._resolveIdle;
			this._resolveIdle = () => {
				existingResolve();
				resolve();
			};
		});
	}
	get size() {
		return this.queue.size;
	}
	get pending() {
		return this._pendingCount;
	}
	get isPaused() {
		return this._isPaused;
	}
}
var pQueue = PQueue$1;

var Manager$2;
var PQueue;
var arm$3;
PQueue = pQueue;
arm$3 = arm;
Manager$2 = class Manager extends arm$3.Manager {
  constructor(PouchDB, concurrency, registry) {
    if (PouchDB == null) {
      throw TypeError('Missing PouchDB');
    }
    super(registry);
    if (concurrency == null) {
      concurrency = 2;
    }
    this.PouchDB = PouchDB;
    this.queue = new PQueue({concurrency});
  }
};
var Manager_1$1 = Manager$2;

var LiveChanges$1;
var Manager$3;
Manager$3 = Manager_1$1;
LiveChanges$1 = class LiveChanges extends Manager$3 {
  register(resource) {
    if (resource.queue === 'live') {
      return this.subscribe(resource);
    }
  }
  subscribe(resource) {
    var local, opts, other, ref, remote;
    ({
      remote,
      identifier: local,
      opts
    } = resource);
    other = (ref = this.registry.resources[local]) != null ? ref.find(function(r) {
      return r.changes && r.queue === 'live';
    }) : void 0;
    if (other) {
      return resource.changes = other.changes;
    }
    return new Promise((resolve, reject) => {
      return this.queue.add(() => {
        opts = Object.assign({
          since: 'now'
        }, opts, {
          live: true,
          include_docs: true
        });
        if (resource.seq) {
          opts.since = resource.seq;
        }
        resource.changes = (new this.PouchDB(remote)).changes(opts);
        resource.changes.on('change', (info) => {
          return this.emit('change', info, local, remote, 'live');
        });
        resource.changes.on('complete', (info) => {
          return this.emit('complete', info, local, remote, 'live');
        });
        resource.changes.on('error', (err) => {
          return this.emit('error', err, local, remote, 'live');
        });
        resolve(resource.changes);
        return resource.changes;
      });
    });
  }
};
var LiveChanges_1 = LiveChanges$1;

var Manager$4;
var Replication$1;
Manager$4 = Manager_1$1;
Replication$1 = class Replication extends Manager$4 {
  register(resource) {
    return this.replicate(resource);
  }
  replicate(resource) {
    var local, opts, remote;
    ({
      remote,
      identifier: local,
      opts
    } = resource);
    return new Promise((resolve, reject) => {
      return this.queue.add(async() => {
        resource.replication = (function() {
          switch (resource.type) {
            case 'push':
              return this.PouchDB.replicate(local, remote, Object.assign({
                since: 0
              }, opts));
            case 'pull':
              return this.PouchDB.replicate(remote, local, Object.assign({
                since: 0
              }, opts));
            case 'sync':
              return this.PouchDB.sync(local, remote, Object.assign({
                since: 0
              }, opts));
            default:
              return Promise.reject(TypeError('Invalid resource type: ' + resource.type));
          }
        }).call(this);
        resource.replication.on('change', (info) => {
          return this.emit('change', info, local, remote, resource.type);
        });
        resource.replication.on('paused', (err) => {
          return this.emit('paused', err, local, remote, resource.type);
        });
        resource.replication.on('active', () => {
          return this.emit('active', local, remote, resource.type);
        });
        resource.replication.on('denied', (err) => {
          return this.emit('denied', err, local, remote, resource.type);
        });
        resource.replication.on('complete', (info) => {
          return this.emit('complete', info, local, remote, resource.type);
        });
        resource.replication.on('error', (err) => {
          return this.emit('error', err, local, remote, resource.type);
        });
        resolve((await resource.replication));
        return resource.replication;
      });
    });
  }
};
var Replication_1 = Replication$1;

var Manager$6;
var PollChanges$1;
Manager$6 = Manager_1$1;
PollChanges$1 = class PollChanges extends Manager$6 {
  register(resource) {
    if (resource.queue === 'poll') {
      return this.subscribe(resource);
    }
  }
  subscribe(resource) {
    var local, opts, other, ref, remote;
    ({
      remote,
      identifier: local,
      opts
    } = resource);
    other = (ref = this.registry.resources[local]) != null ? ref.find(function(r) {
      return r.changes && r.queue === 'poll';
    }) : void 0;
    if (other) {
      return resource.changes = other.changes;
    }
    return new Promise((resolve, reject) => {
      var enqueue;
      opts = Object.assign({
        since: 'now',
        period: 3
      }, opts, {
        live: true,
        include_docs: true
      });
      enqueue = () => {
        if (resource.changes) {
          if (resource.changes.isCancelled) {
            return resolve(resource.changes);
          }
          resource._restart = true;
          resource.changes.cancel();
        }
        return this.queue.add(() => {
          if (resource.seq) {
            opts.since = resource.seq;
          }
          resource.changes = (new this.PouchDB(remote)).changes(opts);
          resource.changes.on('change', (info) => {
            resource.seq = info.seqs;
            return this.emit('change', info, local, remote, 'poll');
          });
          resource.changes.on('complete', (info) => {
            if (!resource._restart) {
              return this.emit('complete', info, local, remote, 'poll');
            }
            return resource._restart = void 0;
          });
          resource.changes.on('error', (err) => {
            return this.emit('error', err, local, remote, 'poll');
          });
          setTimeout(enqueue, opts.period * 1000);
          return resource.changes;
        });
      };
      return enqueue();
    });
  }
};
var PollChanges_1 = PollChanges$1;

var LiveChanges$2;
var Manager$5;
var PollChanges;
var Pouch$1;
var Replication$2;
Manager$5 = Manager_1$1;
LiveChanges$2 = LiveChanges_1;
PollChanges = PollChanges_1;
Replication$2 = Replication_1;
Pouch$1 = class Pouch extends Manager$5 {
  constructor(PouchDB, registry) {
    var concurrency;
    concurrency = 2;
    super(PouchDB, concurrency, registry);
    this.liveChanges = new LiveChanges$2(PouchDB, concurrency, this.registry);
    this.pollChanges = new PollChanges(PouchDB, concurrency, this.registry);
    this.replication = new Replication$2(PouchDB, concurrency, this.registry);
    this.liveChanges.on('change', (...args) => {
      return this.onChange(...args);
    });
    this.liveChanges.on('complete', (...args) => {
      return this.emit('complete', ...args);
    });
    this.liveChanges.on('error', (...args) => {
      return this.emit('error', ...args);
    });
    this.pollChanges.on('change', (...args) => {
      return this.onChange(...args);
    });
    this.pollChanges.on('complete', (...args) => {
      return this.emit('complete', ...args);
    });
    this.pollChanges.on('error', (...args) => {
      return this.emit('error', ...args);
    });
    this.replication.on('change', (...args) => {
      return this.onChange(...args);
    });
    this.replication.on('paused', (...args) => {
      return this.emit('paused', ...args);
    });
    this.replication.on('active', (...args) => {
      return this.emit('active', ...args);
    });
    this.replication.on('denied', (...args) => {
      return this.emit('denied', ...args);
    });
    this.replication.on('complete', (...args) => {
      return this.emit('complete', ...args);
    });
    this.replication.on('error', (...args) => {
      return this.emit('error', ...args);
    });
  }
  async onChange(info, local, remote, mod) {
    var _rev, doc, ref, ref1;
    if (!(_rev = (ref = info.changes) != null ? (ref1 = ref[0]) != null ? ref1.rev : void 0 : void 0)) {
      return;
    }
    try {
      doc = (await (new PouchDB(local)).get(info.id));
      if (doc && doc._rev === _rev) {
        return;
      }
      if (!doc && info.deleted) {
        return;
      }
    } catch (error) {}
    return this.emit('change', info, local, remote, mod);
  }
};
var Pouch_1 = Pouch$1;

var LiveChanges;
var Manager;
var Pouch;
var Registry;
var Replication;
var Resource;
({Registry} = arm);
Resource = Resource_1$1;
Manager = Manager_1$1;
LiveChanges = LiveChanges_1;
Replication = Replication_1;
Pouch = Pouch_1;
var build = {Resource, Registry, Manager, LiveChanges, Replication, Pouch};

module.exports = build;
